/*
 *  Twidere X
 *
 *  Copyright (C) 2020-2021 Tlaster <tlaster@outlook.com>
 * 
 *  This file is part of Twidere X.
 * 
 *  Twidere X is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  Twidere X is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Twidere X. If not, see <http://www.gnu.org/licenses/>.
 */
package com.twidere.twiderex.model

import android.content.Context
import androidx.datastore.core.DataStore
import androidx.datastore.preferences.core.PreferenceDataStoreFactory
import androidx.datastore.preferences.core.Preferences
import androidx.datastore.preferences.core.booleanPreferencesKey
import androidx.datastore.preferences.core.edit
import androidx.datastore.preferences.core.stringPreferencesKey
import androidx.datastore.preferences.preferencesDataStoreFile
import com.twidere.twiderex.scenes.home.HomeMenus
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.flow.map

class AccountPreferences(
    private val dataStore: DataStore<Preferences>,
) {
    private val isNotificationEnabledKey = booleanPreferencesKey("isNotificationEnabled")
    val isNotificationEnabled
        get() = dataStore.data.map { preferences ->
            preferences[isNotificationEnabledKey] ?: true
        }
    val homeMenuOrder
        get() = dataStore.data.map { preferences ->
            if (!preferences.contains(homeMenuOrderKey) || !preferences.contains(visibleHomeMenuKey)) {
                HomeMenus.values().map { it to it.showDefault }
            } else {
                val order = preferences[homeMenuOrderKey].orEmpty()
                    .split(",")
                    .withIndex()
                    .associate { HomeMenus.valueOf(it.value) to it.index }
                val visible = preferences[visibleHomeMenuKey].orEmpty().split(",")
                HomeMenus.values().sortedBy {
                    order[it]
                }.map { it to visible.contains(it.name) }
            }
        }

    suspend fun setIsNotificationEnabled(value: Boolean) {
        dataStore.edit {
            it[isNotificationEnabledKey] = value
        }
    }

    private val homeMenuOrderKey = stringPreferencesKey("homeMenuOrder")
    private val visibleHomeMenuKey = stringPreferencesKey("visibleHomeMenu")
    suspend fun setHomeMenuOrder(
        data: List<Pair<HomeMenus, Boolean>>,
    ) {
        dataStore.edit {
            it[visibleHomeMenuKey] = data.filter { it.second }.joinToString(",") { it.first.name }
        }
        dataStore.edit {
            it[homeMenuOrderKey] = data.joinToString(",") { it.first.name }
        }
    }

    class Factory(
        private val context: Context,
    ) {
        fun create(accountKey: MicroBlogKey) = createAccountPreferences(context, accountKey)

        private fun createAccountPreferences(
            context: Context,
            accountKey: MicroBlogKey,
        ) = AccountPreferences(
            dataStore = PreferenceDataStoreFactory.create(
                corruptionHandler = null,
                migrations = listOf(),
                scope = CoroutineScope(Dispatchers.IO + SupervisorJob())
            ) {
                context.applicationContext.preferencesDataStoreFile(accountKey.toString())
            },
        )
    }
}
